@@ -71,7 +71,7 @@ extern int32_t NSVersionOfRunTimeLibrary(const char* libraryName)            __O
 
 
 /*
- * NSVersionOfRunTimeLibrary() returns the current_version number that the main executable was linked
+ * NSVersionOfLinkTimeLibrary() returns the current_version number that the main executable was linked
  * against at build time.  The libraryName parameter would be "bar" for /path/libbar.3.dylib and
  * "Foo" for /path/Foo.framework/Versions/A/Foo.  It returns -1 if the main executable did not link
  * against the specified library.
@@ -81,8 +81,9 @@ extern int32_t NSVersionOfLinkTimeLibrary(const char* libraryName)           __O
 
 /*
  * _NSGetExecutablePath() copies the path of the main executable into the buffer. The bufsize parameter
- * should initially be the size of the buffer.  The function returns 0 if the path was successfully copied.
- * It returns -1 if the buffer is not large enough, and *bufsize is set to the size required. 
+ * should initially be the size of the buffer.  The function returns 0 if the path was successfully copied,
+ * and *bufsize is left unchanged. It returns -1 if the buffer is not large enough, and *bufsize is set 
+ * to the size required. 
  * 
  * Note that _NSGetExecutablePath will return "a path" to the executable not a "real path" to the executable. 
  * That is the path may be a symbolic link and not the real file. With deep directories the total bufsize 
@@ -93,11 +94,9 @@ extern int _NSGetExecutablePath(char* buf, uint32_t* bufsize)                 __
 
 
 /*
- * _dyld_moninit() and _dyld_func_lookup() are private interface between 
- * dyld and libSystem.
-*/
+ * _dyld_moninit() is a private interface between dyld and libSystem.
+ */
 extern void _dyld_moninit(void (*monaddition)(char *lowpc, char *highpc))    __OSX_AVAILABLE_STARTING(__MAC_10_1, __IPHONE_2_0);
-extern int _dyld_func_lookup(const char* dyld_func_name, void **address)     __OSX_AVAILABLE_BUT_DEPRECATED(__MAC_10_1,__MAC_10_6,__IPHONE_NA,__IPHONE_NA);
 
 
 
