@@ -1,6 +1,6 @@
 /* -*- mode: C++; c-basic-offset: 4; tab-width: 4 -*-
  *
- * Copyright (c) 2004-2008 Apple Inc. All rights reserved.
+ * Copyright (c) 2004-2010 Apple Inc. All rights reserved.
  *
  * @APPLE_LICENSE_HEADER_START@
  * 
@@ -44,9 +44,6 @@
 #include <libkern/OSAtomic.h>
 #include <libkern/OSCacheControl.h>
 
-#if __ppc__ || __ppc64__
-	#include <mach-o/ppc/reloc.h>
-#endif
 #if __x86_64__
 	#include <mach-o/x86_64/reloc.h>
 #endif
@@ -57,14 +54,6 @@
 #include "ImageLoaderMachOClassic.h"
 #include "mach-o/dyld_images.h"
 
-// optimize strcmp for ppc
-#if __ppc__
-	#include <ppc_intrinsics.h>
-#else
-	#define astrcmp(a,b) strcmp(a,b)
-#endif
-
-
 // in dyldStartup.s
 extern "C" void fast_stub_binding_helper_interface();
 
@@ -106,11 +95,12 @@ ImageLoaderMachOClassic* ImageLoaderMachOClassic::instantiateMainExecutable(cons
 	image->setSlide(slide);
 
 	// for PIE record end of program, to know where to start loading dylibs
-	if ( (mh->flags & MH_PIE) && !context.noPIE )
+	if ( slide != 0 )
 		fgNextPIEDylibAddress = (uintptr_t)image->getEnd();
 	
 	image->instantiateFinish(context);
-	
+	image->setMapped(context);
+
 #if __i386__
 	// kernel may have mapped in __IMPORT segment read-only, we need it read/write to do binding
 	if ( image->fReadOnlyImportSegment ) {
@@ -138,43 +128,47 @@ ImageLoaderMachOClassic* ImageLoaderMachOClassic::instantiateMainExecutable(cons
 // create image by mapping in a mach-o file
 ImageLoaderMachOClassic* ImageLoaderMachOClassic::instantiateFromFile(const char* path, int fd, const uint8_t* fileData, 
 															uint64_t offsetInFat, uint64_t lenInFat, const struct stat& info, 
-															unsigned int segCount, unsigned int libCount, const LinkContext& context)
+															unsigned int segCount, unsigned int libCount, 
+															const struct linkedit_data_command* codeSigCmd, const LinkContext& context)
 {
 	ImageLoaderMachOClassic* image = ImageLoaderMachOClassic::instantiateStart((macho_header*)fileData, path, segCount, libCount);
 	try {
 		// record info about file  
 		image->setFileInfo(info.st_dev, info.st_ino, info.st_mtime);
 
+		// if this image is code signed, let kernel validate signature before mapping any pages from image
+		image->loadCodeSignature(codeSigCmd, fd, offsetInFat, context);
+		
 		// mmap segments
 		image->mapSegmentsClassic(fd, offsetInFat, lenInFat, info.st_size, context);
 
-	#if CODESIGNING_SUPPORT
-		// if this code is signed, validate the signature before accessing any mapped pages
-		image->loadCodeSignature(fileData, fd, offsetInFat);
-	#endif
-		
+		// finish up
+		image->instantiateFinish(context);
+
 		// if path happens to be same as in LC_DYLIB_ID load command use that, otherwise malloc a copy of the path
 		const char* installName = image->getInstallPath();
 		if ( (installName != NULL) && (strcmp(installName, path) == 0) && (path[0] == '/') )
 			image->setPathUnowned(installName);
-		else if ( path[0] != '/' ) {
+		else if ( (path[0] != '/') || (strstr(path, "../") != NULL) ) {
+			// rdar://problem/10733082 Fix up @path based paths during introspection
 			// rdar://problem/5135363 turn relative paths into absolute paths so gdb, Symbolication can later find them
 			char realPath[MAXPATHLEN];
-			if ( realpath(path, realPath) != NULL )
-				image->setPath(realPath);
+			if ( fcntl(fd, F_GETPATH, realPath) == 0 ) 
+				image->setPaths(path, realPath);
 			else
 				image->setPath(path);
 		}
 		else 
 			image->setPath(path);
 
+		// make sure path is stable before recording in dyld_all_image_infos
+		image->setMapped(context);
+
 		// pre-fetch content of __DATA segment for faster launches
 		// don't do this on prebound images or if prefetching is disabled
         if ( !context.preFetchDisabled && !image->isPrebindable())
 			image->preFetchDATA(fd, offsetInFat, context);
 
-		// finish up
-		image->instantiateFinish(context);
 	}
 	catch (...) {
 		// ImageLoader::setMapped() can throw an exception to block loading of image
@@ -187,7 +181,7 @@ ImageLoaderMachOClassic* ImageLoaderMachOClassic::instantiateFromFile(const char
 }
 
 // create image by using cached mach-o file
-ImageLoaderMachOClassic* ImageLoaderMachOClassic::instantiateFromCache(const macho_header* mh, const char* path, const struct stat& info,
+ImageLoaderMachOClassic* ImageLoaderMachOClassic::instantiateFromCache(const macho_header* mh, const char* path, long slide, const struct stat& info,
 																unsigned int segCount, unsigned int libCount, const LinkContext& context)
 {
 	ImageLoaderMachOClassic* image = ImageLoaderMachOClassic::instantiateStart(mh, path, segCount, libCount);
@@ -208,6 +202,7 @@ ImageLoaderMachOClassic* ImageLoaderMachOClassic::instantiateFromCache(const mac
 		}
 
 		image->instantiateFinish(context);
+		image->setMapped(context);
 	}
 	catch (...) {
 		// ImageLoader::setMapped() can throw an exception to block loading of image
@@ -240,6 +235,7 @@ ImageLoaderMachOClassic* ImageLoaderMachOClassic::instantiateFromMemory(const ch
 			image->setPath(moduleName);
 
 		image->instantiateFinish(context);
+		image->setMapped(context);
 	}
 	catch (...) {
 		// ImageLoader::setMapped() can throw an exception to block loading of image
@@ -278,9 +274,6 @@ void ImageLoaderMachOClassic::instantiateFinish(const LinkContext& context)
 {
 	// now that segments are mapped in, get real fMachOData, fLinkEditBase, and fSlide
 	this->parseLoadCmds();
-	
-	// notify state change
-	this->setMapped(context);
 }
 
 ImageLoaderMachOClassic::~ImageLoaderMachOClassic()
@@ -298,8 +291,8 @@ uint32_t* ImageLoaderMachOClassic::segmentCommandOffsets() const
 ImageLoader* ImageLoaderMachOClassic::libImage(unsigned int libIndex) const
 {
 	const uintptr_t* images = ((uintptr_t*)(((uint8_t*)this) + sizeof(ImageLoaderMachOClassic) + fSegmentsCount*sizeof(uint32_t)));
-	// mask off low bit
-	return (ImageLoader*)(images[libIndex] & (-2));
+	// mask off low bits
+	return (ImageLoader*)(images[libIndex] & (-4));
 }
 
 bool ImageLoaderMachOClassic::libReExported(unsigned int libIndex) const
@@ -309,13 +302,22 @@ bool ImageLoaderMachOClassic::libReExported(unsigned int libIndex) const
 	return ((images[libIndex] & 1) != 0);
 }	
 
+bool ImageLoaderMachOClassic::libIsUpward(unsigned int libIndex) const
+{
+	const uintptr_t* images = ((uintptr_t*)(((uint8_t*)this) + sizeof(ImageLoaderMachOClassic) + fSegmentsCount*sizeof(uint32_t)));
+	// upward flag is second bit
+	return ((images[libIndex] & 2) != 0);
+}	
 
-void ImageLoaderMachOClassic::setLibImage(unsigned int libIndex, ImageLoader* image, bool reExported)
+
+void ImageLoaderMachOClassic::setLibImage(unsigned int libIndex, ImageLoader* image, bool reExported, bool upward)
 {
 	uintptr_t* images = ((uintptr_t*)(((uint8_t*)this) + sizeof(ImageLoaderMachOClassic) + fSegmentsCount*sizeof(uint32_t)));
 	uintptr_t value = (uintptr_t)image;
 	if ( reExported ) 
 		value |= 1;
+	if ( upward ) 
+		value |= 2;
 	images[libIndex] = value;
 }
 
@@ -506,9 +508,12 @@ void ImageLoaderMachOClassic::mapSegmentsClassic(int fd, uint64_t offsetInFat, u
 		return ImageLoaderMachO::mapSegments(fd, offsetInFat, lenInFat, fileLen, context);
 
 #if SPLIT_SEG_SHARED_REGION_SUPPORT	
-	// try to map into shared region at preferred address
-	if ( mapSplitSegDylibInfoSharedRegion(fd, offsetInFat, lenInFat, fileLen, context) == 0) 
-		return;
+	// don't map split-seg dylibs into shared region if shared cache is in use
+	if ( ! context.dyldLoadedAtSameAddressNeededBySharedCache ) {
+		// try to map into shared region at preferred address
+		if ( mapSplitSegDylibInfoSharedRegion(fd, offsetInFat, lenInFat, fileLen, context) == 0) 
+			return;
+	}
 	// if there is a problem, fall into case where we map file somewhere outside the shared region
 #endif
 
@@ -711,33 +716,6 @@ uintptr_t ImageLoaderMachOClassic::getRelocBase()
 }
 
 
-#if __ppc__
-static inline void otherRelocsPPC(uintptr_t* locationToFix, uint8_t relocationType, uint16_t otherHalf, uintptr_t slide)
-{
-	// low 16 bits of 32-bit ppc instructions need fixing
-	struct ppcInstruction { uint16_t opcode; int16_t immediateValue; };
-	ppcInstruction* instruction = (ppcInstruction*)locationToFix;
-	//uint32_t before = *((uint32_t*)locationToFix);
-	switch ( relocationType )
-	{
-		case PPC_RELOC_LO16: 
-			instruction->immediateValue = ((otherHalf << 16) | instruction->immediateValue) + slide;
-			break;
-		case PPC_RELOC_HI16: 
-			instruction->immediateValue = ((((instruction->immediateValue << 16) | otherHalf) + slide) >> 16);
-			break;
-		case PPC_RELOC_HA16: 
-			int16_t signedOtherHalf = (int16_t)(otherHalf & 0xffff);
-			uint32_t temp = (instruction->immediateValue << 16) + signedOtherHalf + slide;
-			if ( (temp & 0x00008000) != 0 )
-				temp += 0x00008000;
-			instruction->immediateValue = temp >> 16;
-	}
-	//uint32_t after = *((uint32_t*)locationToFix);
-	//dyld::log("dyld: ppc fixup %0p type %d from 0x%08X to 0x%08X\n", locationToFix, relocationType, before, after);
-}
-#endif
-
 #if PREBOUND_IMAGE_SUPPORT
 void ImageLoaderMachOClassic::resetPreboundLazyPointers(const LinkContext& context)
 {
@@ -752,11 +730,6 @@ void ImageLoaderMachOClassic::resetPreboundLazyPointers(const LinkContext& conte
 			if (sreloc->r_length == RELOC_SIZE) {
 				uintptr_t* locationToFix = (uintptr_t*)(sreloc->r_address + relocBase);
 				switch(sreloc->r_type) {
-		#if __ppc__ 
-					case PPC_RELOC_PB_LA_PTR:
-						*locationToFix = sreloc->r_value + slide;
-						break;
-		#endif
 		#if __i386__
 					case GENERIC_RELOC_PB_LA_PTR:
 						*locationToFix = sreloc->r_value + slide;
@@ -779,6 +752,7 @@ void ImageLoaderMachOClassic::resetPreboundLazyPointers(const LinkContext& conte
 
 void ImageLoaderMachOClassic::rebase(const LinkContext& context)
 {
+	CRSetCrashLogMessage2(this->getPath());
 	register const uintptr_t slide = this->fSlide;
 	const uintptr_t relocBase = this->getRelocBase();
 	
@@ -790,6 +764,7 @@ void ImageLoaderMachOClassic::rebase(const LinkContext& context)
 	const relocation_info* const relocsStart = (struct relocation_info*)(&fLinkEditBase[fDynamicInfo->locreloff]);
 	const relocation_info* const relocsEnd = &relocsStart[fDynamicInfo->nlocrel];
 	for (const relocation_info* reloc=relocsStart; reloc < relocsEnd; ++reloc) {
+		uintptr_t rebaseAddr;
 		try {
 	#if LINKEDIT_USAGE_DEBUG
 			noteAccessedLinkEditAddress(reloc);
@@ -804,7 +779,12 @@ void ImageLoaderMachOClassic::rebase(const LinkContext& context)
 				throw "bad local relocation pc_rel";
 			if ( reloc->r_extern != 0 ) 
 				throw "extern relocation found with local relocations";
-			*((uintptr_t*)(reloc->r_address + relocBase)) += slide;
+			rebaseAddr = reloc->r_address + relocBase;
+			if ( ! this->containsAddress((void*)rebaseAddr) )
+				dyld::throwf("local reloc %p not in mapped image\n", (void*)rebaseAddr);
+			*((uintptr_t*)rebaseAddr) += slide;
+			if ( context.verboseRebase )
+				dyld::log("dyld: rebase: %s:*0x%08lX += 0x%08lX\n", this->getShortName(), rebaseAddr, slide);
 		#else	
 			if ( (reloc->r_address & R_SCATTERED) == 0 ) {
 				if ( reloc->r_symbolnum == R_ABS ) {
@@ -813,17 +793,13 @@ void ImageLoaderMachOClassic::rebase(const LinkContext& context)
 				else if (reloc->r_length == RELOC_SIZE) {
 					switch(reloc->r_type) {
 						case GENERIC_RELOC_VANILLA:
-							*((uintptr_t*)(reloc->r_address + relocBase)) += slide;
+							rebaseAddr = reloc->r_address + relocBase;
+							if ( ! this->containsAddress((void*)rebaseAddr) )
+								dyld::throwf("local reloc %p not in mapped image\n", (void*)rebaseAddr);
+							*((uintptr_t*)rebaseAddr) += slide;
+							if ( context.verboseRebase )
+								dyld::log("dyld: rebase: %s:*0x%08lX += 0x%08lX\n", this->getShortName(), rebaseAddr, slide);
 							break;
-			#if __ppc__
-						case PPC_RELOC_HI16: 
-						case PPC_RELOC_LO16: 
-						case PPC_RELOC_HA16: 
-							// some tools leave object file relocations in linked images
-							otherRelocsPPC((uintptr_t*)(reloc->r_address + relocBase), reloc->r_type, reloc[1].r_address, slide);
-							++reloc; // these relocations come in pairs, skip next
-							break;
-			#endif
 						default:
 							throw "unknown local relocation type";
 					}
@@ -838,26 +814,13 @@ void ImageLoaderMachOClassic::rebase(const LinkContext& context)
 					uintptr_t* locationToFix = (uintptr_t*)(sreloc->r_address + relocBase);
 					switch(sreloc->r_type) {
 						case GENERIC_RELOC_VANILLA:
+							if ( ! this->containsAddress((void*)locationToFix) ) 
+								dyld::throwf("local scattered reloc %p not in mapped image\n", locationToFix);
 							*locationToFix += slide;
+							if ( context.verboseRebase )
+								dyld::log("dyld: rebase: %s:*0x%08lX += 0x%08lX\n", this->getShortName(), (uintptr_t)locationToFix, slide);
 							break;
-			#if __ppc__
-						case PPC_RELOC_HI16: 
-						case PPC_RELOC_LO16: 
-						case PPC_RELOC_HA16: 
-							// Metrowerks compiler sometimes leaves object file relocations in linked images???
-							++reloc; // these relocations come in pairs, get next one
-							otherRelocsPPC(locationToFix, sreloc->r_type, reloc->r_address, slide);
-							break;
-						case PPC_RELOC_PB_LA_PTR:
-							// do nothing
-							break;
-			#elif __ppc64__
-						case PPC_RELOC_PB_LA_PTR:
-							// needed for compatibility with ppc64 binaries built with the first ld64
-							// which used PPC_RELOC_PB_LA_PTR relocs instead of GENERIC_RELOC_VANILLA for lazy pointers
-							*locationToFix += slide;
-							break;
-			#elif __i386__
+			#if __i386__
 						case GENERIC_RELOC_PB_LA_PTR:
 							// do nothing
 							break;
@@ -885,6 +848,7 @@ void ImageLoaderMachOClassic::rebase(const LinkContext& context)
 	
 	// update stats
 	fgTotalRebaseFixups += fDynamicInfo->nlocrel;
+	CRSetCrashLogMessage2(NULL);
 }
 
 
@@ -912,7 +876,7 @@ const struct macho_nlist* ImageLoaderMachOClassic::binarySearchWithToc(const cha
 		noteAccessedLinkEditAddress(pivot);
 		noteAccessedLinkEditAddress(pivotStr);
 #endif
-		int cmp = astrcmp(key, pivotStr);
+		int cmp = strcmp(key, pivotStr);
 		if ( cmp == 0 )
 			return pivot;
 		if ( cmp > 0 ) {
@@ -944,7 +908,7 @@ const struct macho_nlist* ImageLoaderMachOClassic::binarySearch(const char* key,
 		noteAccessedLinkEditAddress(pivot);
 		noteAccessedLinkEditAddress(pivotStr);
 #endif
-		int cmp = astrcmp(key, pivotStr);
+		int cmp = strcmp(key, pivotStr);
 		if ( cmp == 0 )
 			return pivot;
 		if ( cmp > 0 ) {
@@ -986,7 +950,7 @@ bool ImageLoaderMachOClassic::containsSymbol(const void* addr) const
 }
 
 
-uintptr_t ImageLoaderMachOClassic::exportedSymbolAddress(const Symbol* symbol) const
+uintptr_t ImageLoaderMachOClassic::exportedSymbolAddress(const LinkContext& context, const Symbol* symbol, const ImageLoader* requestor, bool runResolver) const
 {
 	const struct macho_nlist* sym = (macho_nlist*)symbol;
 	uintptr_t result = sym->n_value + fSlide;
@@ -1066,9 +1030,9 @@ bool ImageLoaderMachOClassic::symbolIsWeakReference(const struct macho_nlist* sy
 	return false;
 }
 
-uintptr_t ImageLoaderMachOClassic::getSymbolAddress(const macho_nlist* sym, const LinkContext& context) const
+uintptr_t ImageLoaderMachOClassic::getSymbolAddress(const macho_nlist* sym, const LinkContext& context, bool runResolver) const
 {
-	return ImageLoaderMachO::getSymbolAddress((Symbol*)sym, this, context);
+	return ImageLoaderMachO::getSymbolAddress((Symbol*)sym, this, context, runResolver);
 }
 
 uintptr_t ImageLoaderMachOClassic::resolveUndefined(const LinkContext& context, const struct macho_nlist* undefinedSymbol, 
@@ -1085,14 +1049,14 @@ uintptr_t ImageLoaderMachOClassic::resolveUndefined(const LinkContext& context,
 		// flat lookup
 		if ( ((undefinedSymbol->n_type & N_PEXT) != 0) && ((undefinedSymbol->n_type & N_TYPE) == N_SECT) ) {
 			// is a multi-module private_extern internal reference that the linker did not optimize away
-			uintptr_t addr = this->getSymbolAddress(undefinedSymbol, context);
+			uintptr_t addr = this->getSymbolAddress(undefinedSymbol, context, false);
 			*foundIn = this;
 			return addr;
 		}
 		const Symbol* sym;
 		if ( context.flatExportFinder(symbolName, &sym, foundIn) ) {
-			if ( (*foundIn != this) && !(*foundIn)->neverUnload() )
-					this->addDynamicReference(*foundIn);
+			if ( *foundIn != this )
+				context.addDynamicReference(this, const_cast<ImageLoader*>(*foundIn));
 			return (*foundIn)->getExportedSymbolAddress(sym, context, this);
 		}
 		// if a bundle is loaded privately the above will not find its exports
@@ -1107,25 +1071,25 @@ uintptr_t ImageLoaderMachOClassic::resolveUndefined(const LinkContext& context,
 			// if reference is weak_import, then it is ok, just return 0
 			return 0;
 		}
-		throwSymbolNotFound(symbolName, this->getPath(), "flat namespace");
+		throwSymbolNotFound(context, symbolName, this->getPath(), "flat namespace");
 	}
 	else {
 		// symbol requires searching images with coalesced symbols (not done during prebinding)
 		if ( !context.prebinding && !dontCoalesce && (symbolIsWeakReference(undefinedSymbol) || symbolIsWeakDefinition(undefinedSymbol)) ) {
 			const Symbol* sym;
 			if ( context.coalescedExportFinder(symbolName, &sym, foundIn) ) {
-				if ( (*foundIn != this) && !(*foundIn)->neverUnload() )
-					this->addDynamicReference(*foundIn);
+				if ( *foundIn != this )
+					context.addDynamicReference(this, const_cast<ImageLoader*>(*foundIn));
 				return (*foundIn)->getExportedSymbolAddress(sym, context, this);
 			}
-			//throwSymbolNotFound(symbolName, this->getPath(), "coalesced namespace");
+			//throwSymbolNotFound(context, symbolName, this->getPath(), "coalesced namespace");
 			//dyld::log("dyld: coalesced symbol %s not found in any coalesced image, falling back to two-level lookup", symbolName);
 		}
 		
 		// if this is a real definition (not an undefined symbol) there is no ordinal
 		if ( (undefinedSymbol->n_type & N_TYPE) == N_SECT ) {
 			// static linker should never generate this case, but if it does, do something sane
-			uintptr_t addr = this->getSymbolAddress(undefinedSymbol, context);
+			uintptr_t addr = this->getSymbolAddress(undefinedSymbol, context, false);
 			*foundIn = this;
 			return addr;
 		}
@@ -1152,7 +1116,7 @@ uintptr_t ImageLoaderMachOClassic::resolveUndefined(const LinkContext& context,
 			if ( context.flatExportFinder(symbolName, &sym, foundIn) )
 				return (*foundIn)->getExportedSymbolAddress(sym, context, this);
 			
-			throwSymbolNotFound(symbolName, this->getPath(), "dynamic lookup");
+			throwSymbolNotFound(context, symbolName, this->getPath(), "dynamic lookup");
 		}
 		else if ( ord <= libraryCount() ) {
 			target = libImage(ord-1);
@@ -1178,7 +1142,7 @@ uintptr_t ImageLoaderMachOClassic::resolveUndefined(const LinkContext& context,
 		else if ( (undefinedSymbol->n_type & N_PEXT) != 0 ) {
 			// don't know why the static linker did not eliminate the internal reference to a private extern definition
 			*foundIn = this;
-			return this->getSymbolAddress(undefinedSymbol, context);
+			return this->getSymbolAddress(undefinedSymbol, context, false);
 		}
 		else if ( (undefinedSymbol->n_desc & N_WEAK_REF) != 0 ) {
 			// if definition not found and reference is weak return 0
@@ -1186,7 +1150,7 @@ uintptr_t ImageLoaderMachOClassic::resolveUndefined(const LinkContext& context,
 		}
 		
 		// nowhere to be found
-		throwSymbolNotFound(symbolName, this->getPath(), target->getPath());
+		throwSymbolNotFound(context, symbolName, this->getPath(), target->getPath());
 	}
 }
 
@@ -1246,6 +1210,8 @@ void ImageLoaderMachOClassic::doBindExternalRelocations(const LinkContext& conte
 					{
 						const struct macho_nlist* undefinedSymbol = &fSymbolTable[reloc->r_symbolnum];
 						uintptr_t* location = ((uintptr_t*)(reloc->r_address + relocBase));
+						if ( ! this->containsAddress((void*)location) )
+							dyld::throwf("external reloc %p not in mapped image %s\n", (void*)location, this->getPath());
 						uintptr_t value = *location;
 						bool symbolAddrCached = true;
 					#if __i386__
@@ -1412,7 +1378,7 @@ uintptr_t ImageLoaderMachOClassic::bindIndirectSymbol(uintptr_t* ptrToBind, cons
 	return targetAddr;
 }
 
-uintptr_t ImageLoaderMachOClassic::doBindFastLazySymbol(uint32_t lazyBindingInfoOffset, const LinkContext& context)
+uintptr_t ImageLoaderMachOClassic::doBindFastLazySymbol(uint32_t lazyBindingInfoOffset, const LinkContext& context, void (*lock)(), void (*unlock)())
 {
 	throw "compressed LINKEDIT lazy binder called with classic LINKEDIT";
 }
@@ -1557,7 +1523,15 @@ uintptr_t ImageLoaderMachOClassic::getAddressCoalIterator(CoalIterator& it, cons
 	}	
 	const struct macho_nlist* sym = &fSymbolTable[symbol_index];
 	//dyld::log("getAddressCoalIterator() => 0x%llX, %s symbol_index=%d, in %s\n", (uint64_t)(sym->n_value + fSlide), &fStrings[sym->n_un.n_strx], symbol_index, this->getPath());
+#if __arm__
+		// processor assumes code address with low bit set is thumb
+		if (sym->n_desc & N_ARM_THUMB_DEF)
+			return (sym->n_value | 1) + fSlide ;
+		else
+			return sym->n_value + fSlide;
+#else
 	return sym->n_value + fSlide;
+#endif
 }
 
 
@@ -1622,7 +1596,7 @@ void ImageLoaderMachOClassic::updateUsesCoalIterator(CoalIterator& it, uintptr_t
 			#if __arm__
 						// if weak and thumb subtract off extra thumb bit
 						if ( (undefinedSymbol->n_desc & N_ARM_THUMB_DEF) != 0 )
-							addend += 1;
+							addend &= -2;
 			#endif
 					}
 				} 
@@ -1645,6 +1619,11 @@ void ImageLoaderMachOClassic::updateUsesCoalIterator(CoalIterator& it, uintptr_t
 					// to be definition address plus addend
 					//dyld::log("weak def, initialValue=0x%lX, undefAddr=0x%lX\n", initialValue, undefinedSymbol->n_value+fSlide);
 					addend = initialValue - (undefinedSymbol->n_value + fSlide);
+		#if __arm__
+					// if weak and thumb subtract off extra thumb bit
+					if ( (undefinedSymbol->n_desc & N_ARM_THUMB_DEF) != 0 )
+						addend &= -2;
+		#endif
 				}
 				else {
 					// nothing fixed up yet, addend is just initial value
@@ -1704,8 +1683,9 @@ void ImageLoaderMachOClassic::updateUsesCoalIterator(CoalIterator& it, uintptr_t
 		}
 		cmd = (const struct load_command*)(((char*)cmd)+cmd->cmdsize);
 	}
-	if ( boundSomething && (targetImage != this) && !targetImage->neverUnload() )
-		this->addDynamicReference(targetImage);
+	if ( boundSomething && (targetImage != this) ) {
+		context.addDynamicReference(this, targetImage);
+	}
 	
 	// mark that this symbol has already been bound, so we don't try to bind again
 	it.type = 1;
@@ -1905,6 +1885,7 @@ void ImageLoaderMachOClassic::initializeLazyStubs(const LinkContext& context)
 
 void ImageLoaderMachOClassic::doBind(const LinkContext& context, bool forceLazysBound)
 {
+	CRSetCrashLogMessage2(this->getPath());
 #if __i386__
 	this->initializeLazyStubs(context);
 #endif
@@ -1918,6 +1899,12 @@ void ImageLoaderMachOClassic::doBind(const LinkContext& context, bool forceLazys
 		// no valid prebinding, so bind symbols.
 		// values bound by name are stored two different ways in classic mach-o:
 		
+	#if TEXT_RELOC_SUPPORT
+		// if there are __TEXT fixups, temporarily make __TEXT writable
+		if ( fTextSegmentBinds ) 
+			this->makeTextSegmentWritable(context, true);
+	#endif
+
 		// 1) external relocations are used for data initialized to external symbols
 		this->doBindExternalRelocations(context);
 		
@@ -1925,10 +1912,17 @@ void ImageLoaderMachOClassic::doBind(const LinkContext& context, bool forceLazys
 		// if this image is in the shared cache, there is no way to reset the lazy pointers, so bind them now
 		this->bindIndirectSymbolPointers(context, true, forceLazysBound || fInSharedCache);
 
+	#if TEXT_RELOC_SUPPORT
+		// if there were __TEXT fixups, restore write protection
+		if ( fTextSegmentBinds ) 
+			this->makeTextSegmentWritable(context, false);
+	#endif	
 	}
 	
 	// set up dyld entry points in image
 	this->setupLazyPointerHandler(context);
+	
+	CRSetCrashLogMessage2(NULL);
 }
 
 void ImageLoaderMachOClassic::doBindJustLazies(const LinkContext& context)
@@ -1937,6 +1931,100 @@ void ImageLoaderMachOClassic::doBindJustLazies(const LinkContext& context)
 	this->bindIndirectSymbolPointers(context, false, true);
 }
 
+void ImageLoaderMachOClassic::doInterpose(const LinkContext& context)
+{
+	if ( context.verboseInterposing )
+		dyld::log("dyld: interposing %lu tuples onto: %s\n", fgInterposingTuples.size(), this->getPath());
+
+	// scan indirect symbols
+	const uint32_t cmd_count = ((macho_header*)fMachOData)->ncmds;
+	const struct load_command* const cmds = (struct load_command*)&fMachOData[sizeof(macho_header)];
+	const struct load_command* cmd = cmds;
+	for (uint32_t i = 0; i < cmd_count; ++i) {
+		switch (cmd->cmd) {
+			case LC_SEGMENT_COMMAND:
+				{
+					const struct macho_segment_command* seg = (struct macho_segment_command*)cmd;
+					const struct macho_section* const sectionsStart = (struct macho_section*)((char*)seg + sizeof(struct macho_segment_command));
+					const struct macho_section* const sectionsEnd = &sectionsStart[seg->nsects];
+					for (const struct macho_section* sect=sectionsStart; sect < sectionsEnd; ++sect) {
+						const uint8_t type = sect->flags & SECTION_TYPE;
+						if ( (type == S_NON_LAZY_SYMBOL_POINTERS) || (type == S_LAZY_SYMBOL_POINTERS) ) {
+							const uint32_t pointerCount = sect->size / sizeof(uintptr_t);
+							uintptr_t* const symbolPointers = (uintptr_t*)(sect->addr + fSlide);
+							for (uint32_t pointerIndex=0; pointerIndex < pointerCount; ++pointerIndex) {
+								for (std::vector<InterposeTuple>::iterator it=fgInterposingTuples.begin(); it != fgInterposingTuples.end(); it++) {
+									// replace all references to 'replacee' with 'replacement'
+									if ( (symbolPointers[pointerIndex] == it->replacee) && (this != it->replacementImage) ) {
+										if ( context.verboseInterposing ) {
+											dyld::log("dyld: interposing: at %p replace 0x%lX with 0x%lX in %s\n", 
+												&symbolPointers[pointerIndex], it->replacee, it->replacement, this->getPath());
+										}
+										symbolPointers[pointerIndex] = it->replacement;
+									}
+								}
+							}
+						}
+				#if __i386__
+						// i386 has special self-modifying stubs that might be prebound to "JMP rel32" that need checking
+						else if ( (type == S_SYMBOL_STUBS) && ((sect->flags & S_ATTR_SELF_MODIFYING_CODE) != 0) && (sect->reserved2 == 5) ) {
+							// check each jmp entry in this section
+							uint8_t* start = (uint8_t*)(sect->addr + this->fSlide);
+							uint8_t* end = start + sect->size;
+							for (uint8_t* entry = start; entry < end; entry += 5) {
+								if ( entry[0] == 0xE9 ) { // 0xE9 == JMP 
+									uint32_t rel32 = *((uint32_t*)&entry[1]); // assume unaligned load of uint32_t is ok
+									uint32_t target = (uint32_t)&entry[5] + rel32;
+									for (std::vector<InterposeTuple>::iterator it=fgInterposingTuples.begin(); it != fgInterposingTuples.end(); it++) {
+										// replace all references to 'replacee' with 'replacement'
+										if ( (it->replacee == target) && (this != it->replacementImage) ) {
+											if ( context.verboseInterposing ) {
+												dyld::log("dyld: interposing: at %p replace JMP 0x%lX with JMP 0x%lX in %s\n", 
+													&entry[1], it->replacee, it->replacement, this->getPath());
+											}
+											uint32_t newRel32 = it->replacement - (uint32_t)&entry[5];
+											*((uint32_t*)&entry[1]) = newRel32; // assume unaligned store of uint32_t is ok
+										}
+									}
+								}
+							}
+						}
+				#endif
+					}
+				}
+				break;
+		}
+		cmd = (const struct load_command*)(((char*)cmd)+cmd->cmdsize);
+	}
+	
+	// scan external relocations 
+	const uintptr_t relocBase = this->getRelocBase();
+	const relocation_info* const relocsStart = (struct relocation_info*)(&fLinkEditBase[fDynamicInfo->extreloff]);
+	const relocation_info* const relocsEnd = &relocsStart[fDynamicInfo->nextrel];
+	for (const relocation_info* reloc=relocsStart; reloc < relocsEnd; ++reloc) {
+		if (reloc->r_length == RELOC_SIZE) {
+			switch(reloc->r_type) {
+				case POINTER_RELOC:
+					{
+						uintptr_t* location = ((uintptr_t*)(reloc->r_address + relocBase));
+						for (std::vector<InterposeTuple>::iterator it=fgInterposingTuples.begin(); it != fgInterposingTuples.end(); it++) {
+							// replace all references to 'replacee' with 'replacement'
+							if ( (*location == it->replacee) && (this != it->replacementImage) ) {
+								if ( context.verboseInterposing ) {
+									dyld::log("dyld: interposing: at %p replace 0x%lX with 0x%lX in %s\n", 
+										location, it->replacee, it->replacement, this->getPath());
+								}
+								*location = it->replacement;
+							}
+						}
+					}
+					break;
+			}
+		}
+	}
+}
+
+
 const char* ImageLoaderMachOClassic::findClosestSymbol(const void* addr, const void** closestAddr) const
 {
 	uintptr_t targetAddress = (uintptr_t)addr - fSlide;
@@ -1970,7 +2058,14 @@ const char* ImageLoaderMachOClassic::findClosestSymbol(const void* addr, const v
 		}
 	}
 	if ( bestSymbol != NULL ) {
+#if __arm__
+		if (bestSymbol->n_desc & N_ARM_THUMB_DEF)
+			*closestAddr = (void*)((bestSymbol->n_value | 1) + fSlide);
+		else
+			*closestAddr = (void*)(bestSymbol->n_value + fSlide);
+#else
 		*closestAddr = (void*)(bestSymbol->n_value + fSlide);
+#endif
 		return &fStrings[bestSymbol->n_un.n_strx];
 	}
 	return NULL;
