@@ -1,6 +1,6 @@
 /* -*- mode: C++; c-basic-offset: 4; tab-width: 4 -*-
  *
- * Copyright (c) 2004-2009 Apple Inc. All rights reserved.
+ * Copyright (c) 2004-2012 Apple Inc. All rights reserved.
  *
  * @APPLE_LICENSE_HEADER_START@
  * 
@@ -25,18 +25,43 @@
 #include <stddef.h>
 #include <string.h>
 #include <malloc/malloc.h>
+#include <sys/mman.h>
 
 #include <crt_externs.h>
 #include <Availability.h>
+#include <vproc_priv.h>
 
 #include "mach-o/dyld.h"
 #include "mach-o/dyld_priv.h"
 
 #include "dyldLock.h"
+#include "start_glue.h"
 
-extern "C" int __cxa_atexit(void (*func)(void *), void *arg, void *dso);
+extern "C" int  __cxa_atexit(void (*func)(void *), void *arg, void *dso);
+extern "C" void __cxa_finalize(const void *dso);
 
-#define DYLD_SHARED_CACHE_SUPPORT (__ppc__ || __i386__ || __ppc64__ || __x86_64__)
+
+#ifndef LC_VERSION_MIN_MACOSX
+	#define LC_VERSION_MIN_MACOSX 0x24
+	struct version_min_command {
+		uint32_t	cmd;		/* LC_VERSION_MIN_MACOSX or
+					   LC_VERSION_MIN_IPHONEOS  */
+		uint32_t	cmdsize;	/* sizeof(struct min_version_command) */
+		uint32_t	version;	/* X.Y.Z is encoded in nibbles xxxx.yy.zz */
+		uint32_t	sdk;		/* X.Y.Z is encoded in nibbles xxxx.yy.zz */
+	};
+#endif
+
+#ifndef LC_VERSION_MIN_IPHONEOS
+	#define LC_VERSION_MIN_IPHONEOS 0x25
+#endif
+
+
+#ifndef LC_LOAD_UPWARD_DYLIB
+	#define	LC_LOAD_UPWARD_DYLIB (0x23|LC_REQ_DYLD)	/* load of dylib whose initializers run later */
+#endif
+
+#define DYLD_SHARED_CACHE_SUPPORT (__i386__ || __x86_64__ || __arm__)
 
 // deprecated APIs are still availble on Mac OS X, but not on iPhone OS
 #if __IPHONE_OS_VERSION_MIN_REQUIRED	
@@ -104,15 +129,16 @@ void NSInstallLinkEditErrorHandlers(
 const NSLinkEditErrorHandlers* handlers)
 {
 	DYLD_LOCK_THIS_BLOCK;
-    static void (*p)(
-	void     (*undefined)(const char* symbol_name),
-	NSModule (*multiple)(NSSymbol s, NSModule old, NSModule newhandler),
-	void     (*linkEdit)(NSLinkEditErrors c, int errorNumber,
-		     const char* fileName, const char* errorString)) = NULL;
+	typedef void (*ucallback_t)(const char* symbol_name);
+ 	typedef NSModule (*mcallback_t)(NSSymbol s, NSModule old, NSModule newhandler);
+	typedef void (*lcallback_t)(NSLinkEditErrors c, int errorNumber,
+								const char* fileName, const char* errorString);
+	static void (*p)(ucallback_t undefined, mcallback_t multiple, lcallback_t linkEdit) = NULL;
 
 	if(p == NULL)
 	    _dyld_func_lookup("__dyld_install_handlers", (void**)&p);
-	p(handlers->undefined, handlers->multiple, handlers->linkEdit);
+	mcallback_t m = handlers->multiple;
+	p(handlers->undefined, m, handlers->linkEdit);
 }
 
 const char* 
@@ -346,6 +372,7 @@ const char* libraryName)
 		switch ( lc->cmd ) { 
 			case LC_LOAD_DYLIB:
 			case LC_LOAD_WEAK_DYLIB:
+			case LC_LOAD_UPWARD_DYLIB:
 				dl = (struct dylib_command *)lc;
 				install_name = (char *)dl + dl->dylib.name.offset;
 				if(names_match(install_name, libraryName) == TRUE)
@@ -399,6 +426,177 @@ const char* libraryName)
 	return(-1);
 }
 
+#define PACKED_VERSION(major, minor, tiny) ((((major) & 0xffff) << 16) | (((minor) & 0xff) << 8) | ((tiny) & 0xff))
+
+
+/*
+ * Returns the sdk version (encode as nibble XXXX.YY.ZZ) the
+ * specified binary was built against.
+ *
+ * First looks for LC_VERSION_MIN_MACOSX/LC_VERSION_MIN_IPHONEOS
+ * in binary and if sdk field is not zero, return that value.
+ * Otherwise, looks for the libSystem.B.dylib the binary linked
+ * against and uses a table to convert that to an sdk version.
+ */
+uint32_t dyld_get_sdk_version(const mach_header* mh)
+{
+	const load_command* cmds = NULL;
+	if ( mh->magic == MH_MAGIC_64 )
+		cmds = (load_command*)((char *)mh + sizeof(mach_header_64));
+	else if ( mh->magic == MH_MAGIC )
+		cmds = (load_command*)((char *)mh + sizeof(mach_header));
+	else
+		return 0;  // not a mach-o file, or wrong endianness
+		
+	const version_min_command* versCmd;
+	const dylib_command* dylibCmd;
+	const load_command* cmd = cmds;
+	const char* dylibName;
+#if __IPHONE_OS_VERSION_MIN_REQUIRED 
+	uint32_t foundationVers = 0;
+#else
+	uint32_t libSystemVers = 0;
+#endif
+	for(uint32_t i = 0; i < mh->ncmds; ++i) {
+		switch ( cmd->cmd ) { 
+#if __IPHONE_OS_VERSION_MIN_REQUIRED 
+			case LC_VERSION_MIN_IPHONEOS:
+#else
+			case LC_VERSION_MIN_MACOSX:
+#endif
+				versCmd = (version_min_command*)cmd;
+#ifdef DICE_KIND_DATA
+				if ( versCmd->sdk != 0 )
+					return versCmd->sdk;	// found explicit SDK version
+#else
+				if ( versCmd->reserved != 0 )
+					return versCmd->reserved;	// found explicit SDK version
+#endif
+				break;
+			case LC_LOAD_DYLIB:
+			case LC_LOAD_WEAK_DYLIB:
+			case LC_LOAD_UPWARD_DYLIB:
+				dylibCmd = (dylib_command*)cmd;
+				dylibName = (char*)dylibCmd + dylibCmd->dylib.name.offset;
+#if __IPHONE_OS_VERSION_MIN_REQUIRED          
+				if ( strcmp(dylibName, "/System/Library/Frameworks/Foundation.framework/Foundation") == 0 )
+					foundationVers = dylibCmd->dylib.current_version;
+#else
+				if ( strcmp(dylibName, "/usr/lib/libSystem.B.dylib") == 0 )
+					libSystemVers = dylibCmd->dylib.current_version;
+#endif
+				break;
+		}
+	    cmd = (load_command*)((char *)cmd + cmd->cmdsize);
+	}
+
+	struct DylibToOSMapping {
+		uint32_t dylibVersion;
+		uint32_t osVersion;
+	};
+	
+#if __IPHONE_OS_VERSION_MIN_REQUIRED
+	static const DylibToOSMapping foundationMapping[] = {
+		{ PACKED_VERSION(678,24,0), DYLD_IOS_VERSION_2_0 },
+		{ PACKED_VERSION(678,26,0), DYLD_IOS_VERSION_2_1 },
+		{ PACKED_VERSION(678,29,0), DYLD_IOS_VERSION_2_2 },
+		{ PACKED_VERSION(678,47,0), DYLD_IOS_VERSION_3_0 },
+		{ PACKED_VERSION(678,51,0), DYLD_IOS_VERSION_3_1 },
+		{ PACKED_VERSION(678,60,0), DYLD_IOS_VERSION_3_2 },
+		{ PACKED_VERSION(751,32,0), DYLD_IOS_VERSION_4_0 },
+		{ PACKED_VERSION(751,37,0), DYLD_IOS_VERSION_4_1 },
+		{ PACKED_VERSION(751,49,0), DYLD_IOS_VERSION_4_2 },
+		{ PACKED_VERSION(751,58,0), DYLD_IOS_VERSION_4_3 },
+		{ PACKED_VERSION(881,0,0),  DYLD_IOS_VERSION_5_0 },
+		{ PACKED_VERSION(890,1,0),  DYLD_IOS_VERSION_5_1 },
+		{ PACKED_VERSION(992,0,0),  DYLD_IOS_VERSION_6_0 },
+		{ PACKED_VERSION(993,0,0),  DYLD_IOS_VERSION_6_1 },  
+		{ PACKED_VERSION(1038,14,0),DYLD_IOS_VERSION_7_0 }, // check final
+		{ PACKED_VERSION(0,0,0),    DYLD_IOS_VERSION_7_0 } 
+	};
+
+	if ( foundationVers != 0 ) {
+		uint32_t lastOsVersion = 0;
+		for (const DylibToOSMapping* p=foundationMapping; ; ++p) {
+			if ( p->dylibVersion == 0 )
+				return p->osVersion;
+			if ( foundationVers < p->dylibVersion )
+				return lastOsVersion;
+			lastOsVersion = p->osVersion;
+		}
+	}
+
+#else
+	// Note: versions are for the GM release.  The last entry should
+	// always be zero.  At the start of the next major version,
+	// a new last entry needs to be added and the previous zero
+	// updated to the GM dylib version.
+	static const DylibToOSMapping libSystemMapping[] = {
+		{ PACKED_VERSION(88,1,3),   DYLD_MACOSX_VERSION_10_4 },
+		{ PACKED_VERSION(111,0,0),  DYLD_MACOSX_VERSION_10_5 },
+		{ PACKED_VERSION(123,0,0),  DYLD_MACOSX_VERSION_10_6 },
+		{ PACKED_VERSION(159,0,0),  DYLD_MACOSX_VERSION_10_7 },
+		{ PACKED_VERSION(169,3,0),  DYLD_MACOSX_VERSION_10_8 },
+		{ PACKED_VERSION(1197,0,0), DYLD_MACOSX_VERSION_10_9 },
+		{ PACKED_VERSION(0,0,0),    DYLD_MACOSX_VERSION_10_9 }
+	};
+
+	if ( libSystemVers != 0 ) {
+		uint32_t lastOsVersion = 0;
+		for (const DylibToOSMapping* p=libSystemMapping; ; ++p) {
+			if ( p->dylibVersion == 0 )
+				return p->osVersion;
+			if ( libSystemVers < p->dylibVersion )
+				return lastOsVersion;
+			lastOsVersion = p->osVersion;
+		}
+	}
+#endif
+	
+	return 0;
+}
+
+uint32_t dyld_get_program_sdk_version()
+{
+	return dyld_get_sdk_version((mach_header*)_NSGetMachExecuteHeader());
+}
+
+
+uint32_t dyld_get_min_os_version(const struct mach_header* mh)
+{
+	const load_command* cmds = NULL;
+	if ( mh->magic == MH_MAGIC_64 )
+		cmds = (load_command*)((char *)mh + sizeof(mach_header_64));
+	else if ( mh->magic == MH_MAGIC )
+		cmds = (load_command*)((char *)mh + sizeof(mach_header));
+	else
+		return 0;  // not a mach-o file, or wrong endianness
+		
+	const version_min_command* versCmd;
+	const load_command* cmd = cmds;
+	for(uint32_t i = 0; i < mh->ncmds; ++i) {
+		switch ( cmd->cmd ) { 
+#if __IPHONE_OS_VERSION_MIN_REQUIRED          
+			case LC_VERSION_MIN_IPHONEOS:
+#else
+			case LC_VERSION_MIN_MACOSX:
+#endif
+				versCmd = (version_min_command*)cmd;
+				return versCmd->version;	// found explicit min OS version
+				break;
+		}
+	    cmd = (load_command*)((char *)cmd + cmd->cmdsize);
+	}
+	return 0;
+}
+
+
+uint32_t dyld_get_program_min_os_version()
+{
+	return dyld_get_min_os_version((mach_header*)_NSGetMachExecuteHeader());
+}
+
+
 #if DEPRECATED_APIS_SUPPORTED
 /*
  * NSCreateObjectFileImageFromFile() creates an NSObjectFileImage for the
@@ -765,7 +963,8 @@ _dyld_register_func_for_add_image(
 void (*func)(const struct mach_header *mh, intptr_t vmaddr_slide))
 {
 	DYLD_LOCK_THIS_BLOCK;
-    static void (*p)(void (*func)(const struct mach_header *mh, intptr_t vmaddr_slide)) = NULL;
+	typedef void (*callback_t)(const struct mach_header *mh, intptr_t vmaddr_slide);
+    static void (*p)(callback_t func) = NULL;
 
 	if(p == NULL)
 	    _dyld_func_lookup("__dyld_register_func_for_add_image", (void**)&p);
@@ -782,7 +981,8 @@ _dyld_register_func_for_remove_image(
 void (*func)(const struct mach_header *mh, intptr_t vmaddr_slide))
 {
 	DYLD_LOCK_THIS_BLOCK;
-    static void (*p)(void (*func)(const struct mach_header *mh, intptr_t vmaddr_slide)) = NULL;
+	typedef void (*callback_t)(const struct mach_header *mh, intptr_t vmaddr_slide);
+    static void (*p)(callback_t func) = NULL;
 
 	if(p == NULL)
 	    _dyld_func_lookup("__dyld_register_func_for_remove_image", (void**)&p);
@@ -972,7 +1172,8 @@ void _dyld_moninit(
 void (*monaddition)(char *lowpc, char *highpc))
 {
 	DYLD_LOCK_THIS_BLOCK;
-    static void (*p)(void (*monaddition)(char *lowpc, char *highpc)) = NULL;
+	typedef void (*monproc)(char *lowpc, char *highpc);
+    static void (*p)(monproc monaddition) = NULL;
 
 	if(p == NULL)
 	    _dyld_func_lookup("__dyld_moninit", (void**)&p);
@@ -1052,6 +1253,32 @@ static char* getPerThreadBufferFor_dlerror(uint32_t sizeRequired)
 	return data->message;
 }
 
+// <rdar://problem/10595338> dlerror buffer leak
+// Only allocate buffer if an actual error message needs to be set
+static bool hasPerThreadBufferFor_dlerror()
+{
+	if (!dlerrorPerThreadKeyInitialized ) 
+		return false;
+		
+	return (pthread_getspecific(dlerrorPerThreadKey) != NULL);
+}
+
+// use non-lazy pointer to vproc_swap_integer so that lazy binding does not recurse
+typedef vproc_err_t (*vswapproc)(vproc_t vp, vproc_gsk_t key,int64_t *inval, int64_t *outval);
+static vswapproc swapProc = &vproc_swap_integer;
+
+static bool isLaunchdOwned() {
+	static bool first = true;
+	static bool result;
+	if ( first ) {
+		int64_t val = 0;
+		(*swapProc)(NULL, VPROC_GSK_IS_MANAGED, NULL, &val);
+		result = ( val != 0 );
+		first = false;
+	}
+	return result;
+}
+
 
 #if DYLD_SHARED_CACHE_SUPPORT
 static void shared_cache_missing()
@@ -1067,7 +1294,7 @@ static void shared_cache_out_of_date()
 
 
 // the table passed to dyld containing thread helpers
-static dyld::LibSystemHelpers sHelpers = { 6, &dyldGlobalLockAcquire, &dyldGlobalLockRelease,  
+static dyld::LibSystemHelpers sHelpers = { 12, &dyldGlobalLockAcquire, &dyldGlobalLockRelease,
 									&getPerThreadBufferFor_dlerror, &malloc, &free, &__cxa_atexit,
 						#if DYLD_SHARED_CACHE_SUPPORT
 									&shared_cache_missing, &shared_cache_out_of_date,
@@ -1076,23 +1303,31 @@ static dyld::LibSystemHelpers sHelpers = { 6, &dyldGlobalLockAcquire, &dyldGloba
 						#endif
 									NULL, NULL,
 									&pthread_key_create, &pthread_setspecific,
-									&malloc_size };
+									&malloc_size,
+									&pthread_getspecific,
+									&__cxa_finalize,
+									address_of_start,
+									&hasPerThreadBufferFor_dlerror,
+									&isLaunchdOwned,
+									&vm_allocate,
+									&mmap};
 
 
 //
 // during initialization of libSystem this routine will run
 // and call dyld, registering the helper functions.
 //
-extern "C" void _dyld_initializer() __attribute__((visibility("hidden")));
+extern "C" void tlv_initializer();
+extern "C" void _dyld_initializer();
 void _dyld_initializer()
-{
-	DYLD_LOCK_INITIALIZER;
-	
+{	
    void (*p)(dyld::LibSystemHelpers*);
 
 	_dyld_func_lookup("__dyld_register_thread_helpers", (void**)&p);
 	if(p != NULL)
 		p(&sHelpers);
+		
+	tlv_initializer();
 }
 
 
@@ -1187,7 +1422,6 @@ const struct dyld_all_image_infos* _dyld_get_all_image_infos()
 }
 
 #if !__arm__
-__attribute__((visibility("hidden"))) 
 bool _dyld_find_unwind_sections(void* addr, dyld_unwind_sections* info)
 {
 	DYLD_NO_LOCK_THIS_BLOCK;
@@ -1200,7 +1434,7 @@ bool _dyld_find_unwind_sections(void* addr, dyld_unwind_sections* info)
 #endif
 
 
-#if __i386__ || __x86_64__
+#if __i386__ || __x86_64__ || __arm__
 __attribute__((visibility("hidden"))) 
 void* _dyld_fast_stub_entry(void* loadercache, long lazyinfo)
 {
@@ -1224,4 +1458,43 @@ const char* dyld_image_path_containing_address(const void* addr)
 	return p(addr);
 }
 
+#if __IPHONE_OS_VERSION_MIN_REQUIRED	
+bool dyld_shared_cache_some_image_overridden()
+{
+	DYLD_NO_LOCK_THIS_BLOCK;
+    static bool (*p)() = NULL;
+
+	if(p == NULL)
+	    _dyld_func_lookup("__dyld_shared_cache_some_image_overridden", (void**)&p);
+	return p();
+}
+#endif
+
+
+bool dyld_process_is_restricted()
+{
+	DYLD_NO_LOCK_THIS_BLOCK;
+    static bool (*p)() = NULL;
+	
+	if(p == NULL)
+	    _dyld_func_lookup("__dyld_process_is_restricted", (void**)&p);
+	return p();
+}
+
+
+
+
+// SPI called __fork
+void _dyld_fork_child()
+{
+	DYLD_NO_LOCK_THIS_BLOCK;
+    static void (*p)() = NULL;
+
+	if(p == NULL)
+	    _dyld_func_lookup("__dyld_fork_child", (void**)&p);
+	return p();
+}
+
+
+
 
